#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.enums as enums
import octobot_commons.constants as constants


def get_eval_time(full_candle=None, time_frame=None, partial_candle=None, kline=None):
    if full_candle is not None and time_frame is not None:
        # add one full time frame seconds since a full candle is available when the next has started
        return full_candle[enums.PriceIndexes.IND_PRICE_TIME.value] + \
               enums.TimeFramesMinutes[enums.TimeFrames(time_frame)] * constants.MINUTE_TO_SECONDS
    if partial_candle is not None:
        return partial_candle[enums.PriceIndexes.IND_PRICE_TIME.value]
    if kline is not None:
        return kline[enums.PriceIndexes.IND_PRICE_TIME.value]
    raise ValueError("Invalid arguments")


def get_shortest_time_frame(ideal_time_frame, preferred_available_time_frames, others):
    if ideal_time_frame in preferred_available_time_frames:
        return ideal_time_frame
    if preferred_available_time_frames:
        return preferred_available_time_frames[0]
    else:
        return others[0]



#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest
import mock

import octobot_tentacles_manager.api as tentacles_api
import octobot_evaluators.evaluators as evaluators
import octobot_commons.enums as enums
import octobot_commons.symbol_util as symbol_util
from tests import event_loop, matrix_id, install_tentacles, evaluators_and_matrix_channels

pytestmark = pytest.mark.asyncio

exchange_name = "TEST_EXCHANGE_NAME"
bot_id = "TEST_BOT_ID"
symbols_by_crypto_currencies = {
    "Bitcoin": ["BTC/USDT"],
    "Ethereum": ["ETH/USD", "ETH/BTC"]
}
symbols = ["BTC/USDT", "ETH/USD", "ETH/BTC"]
time_frames = [enums.TimeFrames.ONE_HOUR, enums.TimeFrames.FOUR_HOURS]

crypto_currency_name_by_crypto_currencies = {}
symbols_by_crypto_currency_tickers = {}
for name, symbol_list in symbols_by_crypto_currencies.items():
    ticker = symbol_util.split_symbol(symbol_list[0])[0]
    crypto_currency_name_by_crypto_currencies[ticker] = name
    symbols_by_crypto_currency_tickers[ticker] = symbol_list


@pytest.mark.usefixtures("event_loop", "install_tentacles")
async def test_create_all_type_evaluators(evaluators_and_matrix_channels):
    tentacles_setup_config = tentacles_api.get_tentacles_setup_config()
    created_evaluators = await evaluators.create_all_type_evaluators(tentacles_setup_config,
                                                                     matrix_id=evaluators_and_matrix_channels,
                                                                     exchange_name=exchange_name,
                                                                     bot_id=bot_id,
                                                                     symbols_by_crypto_currencies=symbols_by_crypto_currencies,
                                                                     symbols=symbols,
                                                                     time_frames=time_frames)

    assert not created_evaluators  # Trading package is not installed


@pytest.mark.usefixtures("event_loop", "install_tentacles")
async def test_create_strategy_evaluators(evaluators_and_matrix_channels):
    import tentacles
    await _test_evaluators_creation(evaluators.StrategyEvaluator, evaluators_and_matrix_channels, [
        tentacles.SimpleStrategyEvaluator,
        tentacles.DipAnalyserStrategyEvaluator,
        tentacles.MoveSignalsStrategyEvaluator
    ])


@pytest.mark.usefixtures("event_loop", "install_tentacles")
async def test_create_ta_evaluators(evaluators_and_matrix_channels):
    import tentacles
    await _test_evaluators_creation(evaluators.TAEvaluator, evaluators_and_matrix_channels, [
        tentacles.RSIMomentumEvaluator,
        tentacles.ADXMomentumEvaluator,
        tentacles.StochasticRSIVolatilityEvaluator
    ])


@pytest.mark.usefixtures("event_loop", "install_tentacles")
async def test_create_social_evaluators(evaluators_and_matrix_channels):
    import tentacles
    await _test_evaluators_creation(evaluators.SocialEvaluator, evaluators_and_matrix_channels, [
        tentacles.RedditForumEvaluator
    ])


@pytest.mark.usefixtures("event_loop", "install_tentacles")
async def test_create_rt_evaluators(evaluators_and_matrix_channels):
    import tentacles
    await _test_evaluators_creation(evaluators.RealTimeEvaluator, evaluators_and_matrix_channels, [
        tentacles.InstantFluctuationsEvaluator
    ])


async def _test_evaluators_creation(evaluator_parent_class, fixture_matrix_id, expected_evaluators):
    tentacles_setup_config = tentacles_api.get_tentacles_setup_config()

    # mock start method to prevent side effects (octobot-trading imports, etc)
    with mock.patch.object(evaluator_parent_class, "start", mock.AsyncMock()):
        created_evaluators = await evaluators.create_evaluators(evaluator_parent_class=evaluator_parent_class,
                                                                tentacles_setup_config=tentacles_setup_config,
                                                                matrix_id=fixture_matrix_id,
                                                                exchange_name=exchange_name,
                                                                bot_id=bot_id,
                                                                crypto_currency_name_by_crypto_currencies=crypto_currency_name_by_crypto_currencies,
                                                                symbols_by_crypto_currency_tickers=symbols_by_crypto_currency_tickers,
                                                                symbols=symbols,
                                                                time_frames=time_frames)
    assert created_evaluators
    assert all([evaluator.__class__ in expected_evaluators for evaluator in created_evaluators])



#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_evaluators.enums
from octobot_evaluators.evaluators import abstract_evaluator
from octobot_evaluators.evaluators.abstract_evaluator import (
    AbstractEvaluator,
)

from octobot_evaluators.evaluators import evaluator_factory
from octobot_evaluators.evaluators import realtime_evaluator
from octobot_evaluators.evaluators import social_evaluator
from octobot_evaluators.evaluators import TA_evaluator
from octobot_evaluators.evaluators import abstract_util
from octobot_evaluators.evaluators import strategy_evaluator

from octobot_evaluators.evaluators.evaluator_factory import (
    create_evaluator,
    create_all_type_evaluators,
    create_evaluators,
)
from octobot_evaluators.evaluators.realtime_evaluator import (
    RealTimeEvaluator,
)
from octobot_evaluators.evaluators.social_evaluator import (
    SocialEvaluator,
)
from octobot_evaluators.evaluators.TA_evaluator import (
    TAEvaluator,
)
from octobot_evaluators.evaluators.abstract_util import (
    AbstractUtil,
)
from octobot_evaluators.evaluators.strategy_evaluator import (
    StrategyEvaluator,
)


EvaluatorClassTypes = {
    octobot_evaluators.enums.EvaluatorMatrixTypes.TA.value: TAEvaluator,
    octobot_evaluators.enums.EvaluatorMatrixTypes.SOCIAL.value: SocialEvaluator,
    octobot_evaluators.enums.EvaluatorMatrixTypes.REAL_TIME.value: RealTimeEvaluator,
    octobot_evaluators.enums.EvaluatorMatrixTypes.STRATEGIES.value: StrategyEvaluator
}

evaluator_class_str_to_matrix_type_dict = {
    "TAEvaluator": octobot_evaluators.enums.EvaluatorMatrixTypes.TA,
    "SocialEvaluator": octobot_evaluators.enums.EvaluatorMatrixTypes.SOCIAL,
    "RealTimeEvaluator": octobot_evaluators.enums.EvaluatorMatrixTypes.REAL_TIME,
    "StrategyEvaluator": octobot_evaluators.enums.EvaluatorMatrixTypes.STRATEGIES
}

__all__ = [
    "RealTimeEvaluator",
    "AbstractEvaluator",
    "SocialEvaluator",
    "TAEvaluator",
    "AbstractUtil",
    "StrategyEvaluator",
    "EvaluatorClassTypes",
    "create_evaluator",
    "create_all_type_evaluators",
    "create_evaluators",
    "evaluator_class_str_to_matrix_type_dict",
]



#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import octobot_commons.constants as common_constants
import octobot_commons.enums as common_enums
import octobot_commons.time_frame_manager as time_frame_manager

import octobot_evaluators.constants as constants
import octobot_evaluators.enums as enums
import octobot_evaluators.evaluators as evaluator
import octobot_evaluators.evaluators.channel as evaluator_channels
import octobot_evaluators.matrix as matrix

import octobot_tentacles_manager.api as api


class StrategyEvaluator(evaluator.AbstractEvaluator):
    __metaclass__ = evaluator.AbstractEvaluator

    def __init__(self):
        super().__init__()
        self.consumer_instance = None
        self.strategy_time_frames = []
        self.evaluations_last_updates = {}
        self.allowed_time_delta = None

        # caches
        self.available_evaluators_cache = {}
        self.available_time_frames_cache = {}
        self.available_node_paths_cache = {}

    async def start(self, bot_id: str) -> bool:
        """
        Default Strategy start: to be overwritten
        Subscribe to Matrix notification from self.symbols and self.time_frames
        :return: success of the evaluator's start
        """
        await super().start(bot_id)
        self.consumer_instance = await evaluator_channels.get_chan(constants.MATRIX_CHANNEL,
                                                                   self.matrix_id).new_consumer(
            self.strategy_matrix_callback,
            priority_level=self.priority_level,
            exchange_name=self.exchange_name if self.exchange_name else common_constants.CHANNEL_WILDCARD)
        self._init_exchange_allowed_time_delta(self.exchange_name, self.matrix_id)
        return True

    async def strategy_completed(self,
                                 cryptocurrency: str = None,
                                 symbol: str = None,
                                 eval_note=None,
                                 eval_time=0,
                                 notify=True) -> None:
        """
        Main async method to notify that a strategy has updated its evaluation
        :param cryptocurrency: evaluated cryptocurrency
        :param symbol: evaluated symbol
        :param eval_note: if None = self.eval_note
        :param eval_time: the time of the evaluation if relevant, default is 0
        :param notify: if true, will trigger matrix consumers
        :return: None
        """
        return await self.evaluation_completed(cryptocurrency=cryptocurrency,
                                               symbol=symbol,
                                               time_frame=None,
                                               eval_note=eval_note,
                                               eval_time=eval_time,
                                               notify=notify,
                                               origin_consumer=self.consumer_instance)

    def is_technical_evaluator_cycle_complete(self, matrix_id, evaluator_name, evaluator_type, exchange_name,
                                              cryptocurrency, symbol, time_frame) -> bool:
        """
        :return: True if the strategy is to be waken up by technical evaluators at the moment of this call.
        This avoids partial time frame updates wakeup.
        Override if necessary
        """
        # 1. Ensure this evaluation has not already been sent
        # 2. Ensure every technical evaluator form this time frame are valid
        return not self._already_sent_this_technical_evaluation(matrix_id,
                                                                evaluator_name,
                                                                evaluator_type,
                                                                exchange_name,
                                                                cryptocurrency,
                                                                symbol,
                                                                time_frame) and \
               self._are_every_evaluation_valid(matrix_id,
                                                evaluator_name,
                                                evaluator_type,
                                                exchange_name,
                                                cryptocurrency,
                                                symbol,
                                                time_frame)

    def clear_cache(self):
        self.available_evaluators_cache = {}
        self.available_time_frames_cache = {}
        self.available_node_paths_cache = {}

    async def stop(self) -> None:
        if self.consumer_instance:
            await evaluator_channels.get_chan(constants.MATRIX_CHANNEL,
                                              self.matrix_id).remove_consumer(self.consumer_instance)

    async def strategy_matrix_callback(self,
                                       matrix_id,
                                       evaluator_name,
                                       evaluator_type,
                                       eval_note,
                                       eval_note_type,
                                       exchange_name,
                                       cryptocurrency,
                                       symbol,
                                       time_frame):
        # if this callback is from a technical evaluator: ensure strategy should be notified at this moment
        if evaluator_type == enums.EvaluatorMatrixTypes.TA.value:
            # ensure this time frame is within the strategy's time frames
            if common_enums.TimeFrames(time_frame) not in self.strategy_time_frames or \
                    not self.is_technical_evaluator_cycle_complete(matrix_id,
                                                                   evaluator_name,
                                                                   evaluator_type,
                                                                   exchange_name,
                                                                   cryptocurrency,
                                                                   symbol,
                                                                   time_frame):
                # do not call the strategy
                return
        await self.matrix_callback(
            matrix_id,
            evaluator_name,
            evaluator_type,
            eval_note,
            eval_note_type,
            exchange_name,
            cryptocurrency,
            symbol,
            time_frame
        )

    async def matrix_callback(self,
                              matrix_id,
                              evaluator_name,
                              evaluator_type,
                              eval_note,
                              eval_note_type,
                              exchange_name,
                              cryptocurrency,
                              symbol,
                              time_frame):
        # To be used to trigger an evaluation
        # Do not forget to check if evaluator_name is self.name
        pass

    def _are_every_evaluation_valid(self,
                                    matrix_id,
                                    evaluator_name,
                                    evaluator_type,
                                    exchange_name,
                                    cryptocurrency,
                                    symbol,
                                    time_frame
                                    ):
        to_validate_node_paths = self._get_available_node_paths(matrix_id,
                                                                evaluator_type,
                                                                exchange_name,
                                                                cryptocurrency,
                                                                symbol,
                                                                use_cache=True)
        current_time = self._get_exchange_current_time(exchange_name, matrix_id)
        # ensure all evaluations are valid (do not trigger on an expired evaluation)
        if all(matrix.is_tentacle_value_valid(self.matrix_id, evaluation_node_path,
                                              timestamp=current_time,
                                              delta=self.allowed_time_delta)
               for evaluation_node_path in to_validate_node_paths):
            self._save_last_evaluation(matrix_id, exchange_name, evaluator_type, evaluator_name,
                                       cryptocurrency, symbol, time_frame)
            return True
        return False

    def _get_available_node_paths(self,
                                  matrix_id,
                                  evaluator_type,
                                  exchange_name,
                                  cryptocurrency,
                                  symbol,
                                  use_cache=True):
        if use_cache:
            try:
                return self.available_node_paths_cache[matrix_id][exchange_name][evaluator_type][cryptocurrency][symbol]
            except KeyError:
                # No cache usage here to be use to refresh data
                node_paths = self._inner_get_available_node_paths(matrix_id, evaluator_type, exchange_name,
                                                                  cryptocurrency, symbol, use_cache=False)
                if matrix_id not in self.available_node_paths_cache:
                    self.available_node_paths_cache[matrix_id] = {}
                if exchange_name not in self.available_node_paths_cache[matrix_id]:
                    self.available_node_paths_cache[matrix_id][exchange_name] = {}
                if evaluator_type not in self.available_node_paths_cache[matrix_id][exchange_name]:
                    self.available_node_paths_cache[matrix_id][exchange_name][evaluator_type] = {}
                if cryptocurrency not in self.available_node_paths_cache[matrix_id][exchange_name][evaluator_type]:
                    self.available_node_paths_cache[matrix_id][exchange_name][evaluator_type][cryptocurrency] = {}
                self.available_node_paths_cache[matrix_id][exchange_name][evaluator_type][cryptocurrency][symbol] = \
                    node_paths
                return node_paths
        return self._inner_get_available_node_paths(matrix_id, evaluator_type, exchange_name,
                                                    cryptocurrency, symbol, use_cache=use_cache)

    def _inner_get_available_node_paths(self, matrix_id, evaluator_type, exchange_name, cryptocurrency, symbol,
                                        use_cache=True):
        return [
            matrix.get_matrix_default_value_path(tentacle_name=evaluator,
                                                 tentacle_type=evaluator_type,
                                                 exchange_name=exchange_name,
                                                 cryptocurrency=cryptocurrency,
                                                 symbol=symbol,
                                                 time_frame=time_frame)
            for time_frame in self.get_available_time_frames(matrix_id, exchange_name, evaluator_type,
                                                             cryptocurrency, symbol, use_cache=use_cache)
            if common_enums.TimeFrames(time_frame) in self.strategy_time_frames
            for evaluator in self._get_available_evaluators(matrix_id, exchange_name, evaluator_type,
                                                            use_cache=use_cache)
        ]

    def _save_last_evaluation(self, matrix_id, exchange_name, evaluator_type, tentacle_name,
                              cryptocurrency, symbol, time_frame):
        self._set_last_evaluation_time(exchange_name,
                                       evaluator_type,
                                       cryptocurrency,
                                       symbol,
                                       time_frame,
                                       matrix.get_tentacle_eval_time(matrix_id,
                                                                     matrix.get_matrix_default_value_path(
                                                                         tentacle_name=tentacle_name,
                                                                         tentacle_type=evaluator_type,
                                                                         exchange_name=exchange_name,
                                                                         cryptocurrency=cryptocurrency,
                                                                         symbol=symbol,
                                                                         time_frame=time_frame)
                                                                     )
                                       )

    def _already_sent_this_technical_evaluation(self, matrix_id, evaluator, evaluator_type, exchange_name,
                                                cryptocurrency, symbol, time_frame):
        try:
            update_time = matrix.get_tentacle_eval_time(matrix_id,
                                                        matrix.get_matrix_default_value_path(
                                                            tentacle_name=evaluator,
                                                            tentacle_type=evaluator_type,
                                                            exchange_name=exchange_name,
                                                            cryptocurrency=cryptocurrency,
                                                            symbol=symbol,
                                                            time_frame=time_frame)
                                                        )
            return self.evaluations_last_updates[exchange_name][evaluator_type][cryptocurrency][symbol][time_frame] \
                   == update_time
        except KeyError:
            return False

    def _set_last_evaluation_time(self, exchange_name, evaluator_type, cryptocurrency, symbol, time_frame, value):
        try:
            self.evaluations_last_updates[exchange_name][evaluator_type][cryptocurrency][symbol][time_frame] = value
        except KeyError:
            if exchange_name not in self.evaluations_last_updates:
                self.evaluations_last_updates[exchange_name] = {}
            if evaluator_type not in self.evaluations_last_updates[exchange_name]:
                self.evaluations_last_updates[exchange_name][evaluator_type] = {}
            if cryptocurrency not in self.evaluations_last_updates[exchange_name][evaluator_type]:
                self.evaluations_last_updates[exchange_name][evaluator_type][cryptocurrency] = {}
            if symbol not in self.evaluations_last_updates[exchange_name][evaluator_type][cryptocurrency]:
                self.evaluations_last_updates[exchange_name][evaluator_type][cryptocurrency][symbol] = {}
            self.evaluations_last_updates[exchange_name][evaluator_type][cryptocurrency][symbol] = {
                time_frame: value
            }

    def _get_exchange_current_time(self, exchange_name, matrix_id):
        try:
            import octobot_trading.api as exchange_api
            exchange_manager = exchange_api.get_exchange_manager_from_exchange_name_and_id(
                exchange_name,
                exchange_api.get_exchange_id_from_matrix_id(exchange_name, matrix_id)
            )
            return exchange_api.get_exchange_current_time(exchange_manager)
        except ImportError:
            self.logger.error("Strategy requires OctoBot-Trading package installed")

    def _init_exchange_allowed_time_delta(self, exchange_name, matrix_id):
        try:
            import octobot_trading.api as exchange_api
            exchange_manager = exchange_api.get_exchange_manager_from_exchange_name_and_id(
                exchange_name,
                exchange_api.get_exchange_id_from_matrix_id(exchange_name, matrix_id)
            )
            self.allowed_time_delta = exchange_api.get_exchange_allowed_time_lag(exchange_manager)
        except ImportError:
            self.logger.error("Strategy requires OctoBot-Trading package installed")

    async def evaluators_callback(self,
                                  matrix_id,
                                  evaluator_name,
                                  evaluator_type,
                                  exchange_name,
                                  cryptocurrency,
                                  symbol,
                                  time_frame,
                                  data):
        # Used to communicate between evaluators
        if data[constants.EVALUATOR_CHANNEL_DATA_ACTION] == constants.RESET_EVALUATION:
            for time_frame in data[constants.EVALUATOR_CHANNEL_DATA_TIME_FRAMES]:
                self._set_last_evaluation_time(exchange_name, enums.EvaluatorMatrixTypes.TA.value,
                                               cryptocurrency, symbol, time_frame.value, None)

    def _get_available_evaluators(self, matrix_id, exchange_name, tentacle_type, use_cache=True):
        if use_cache:
            try:
                return self.available_evaluators_cache[matrix_id][exchange_name][tentacle_type]
            except KeyError:
                available_evaluators = matrix.get_node_children_by_names_at_path(
                    matrix_id, matrix.get_tentacle_path(exchange_name=exchange_name, tentacle_type=tentacle_type)
                ).keys()
                if matrix_id not in self.available_evaluators_cache:
                    self.available_evaluators_cache[matrix_id] = {}
                if exchange_name not in self.available_evaluators_cache[matrix_id]:
                    self.available_evaluators_cache[matrix_id][exchange_name] = {}
                self.available_evaluators_cache[matrix_id][exchange_name][tentacle_type] = available_evaluators
                return available_evaluators
        return matrix.get_node_children_by_names_at_path(
            matrix_id, matrix.get_tentacle_path(exchange_name=exchange_name, tentacle_type=tentacle_type)
        ).keys()

    def get_available_time_frames(self,
                                  matrix_id,
                                  exchange_name=None,
                                  tentacle_type=None,
                                  cryptocurrency=None,
                                  symbol=None,
                                  use_cache=True):
        if use_cache:
            try:
                return self.available_time_frames_cache[matrix_id][exchange_name][tentacle_type][cryptocurrency][symbol]
            except KeyError:
                available_time_frames = matrix.get_available_time_frames(
                    matrix_id, exchange_name, tentacle_type, cryptocurrency, symbol)
                if matrix_id not in self.available_time_frames_cache:
                    self.available_time_frames_cache[matrix_id] = {}
                if exchange_name not in self.available_time_frames_cache[matrix_id]:
                    self.available_time_frames_cache[matrix_id][exchange_name] = {}
                if tentacle_type not in self.available_time_frames_cache[matrix_id][exchange_name]:
                    self.available_time_frames_cache[matrix_id][exchange_name][tentacle_type] = {}
                if cryptocurrency not in self.available_time_frames_cache[matrix_id][exchange_name][tentacle_type]:
                    self.available_time_frames_cache[matrix_id][exchange_name][tentacle_type][cryptocurrency] = {}
                self.available_time_frames_cache[matrix_id][exchange_name][tentacle_type][cryptocurrency][symbol] = \
                    available_time_frames
                return available_time_frames
        return matrix.get_available_time_frames(matrix_id, exchange_name, tentacle_type, cryptocurrency, symbol)

    def _get_tentacle_registration_topic(self, all_symbols_by_crypto_currencies, time_frames, real_time_time_frames):
        strategy_currencies, symbols, self.strategy_time_frames = super()._get_tentacle_registration_topic(
            all_symbols_by_crypto_currencies,
            time_frames,
            real_time_time_frames)
        self.all_symbols_by_crypto_currencies = all_symbols_by_crypto_currencies
        # by default no time frame registration for strategies
        return strategy_currencies, symbols, [self.time_frame]

    @classmethod
    def get_required_time_frames(cls, config: dict):
        if constants.CONFIG_FORCED_TIME_FRAME in config:
            return time_frame_manager.parse_time_frames(config[constants.CONFIG_FORCED_TIME_FRAME])
        strategy_config: dict = api.get_tentacle_config(cls)
        if constants.STRATEGIES_REQUIRED_TIME_FRAME in strategy_config:
            return time_frame_manager.parse_time_frames(strategy_config[constants.STRATEGIES_REQUIRED_TIME_FRAME])
        else:
            raise Exception(f"'{constants.STRATEGIES_REQUIRED_TIME_FRAME}' is missing in configuration file")

    @classmethod
    def get_required_evaluators(cls, strategy_config: dict = None) -> list:
        """
        :param strategy_config: the strategy configuration dict
        :return: the list of required evaluators, [CONFIG_WILDCARD] means any evaluator
        """
        strategy_config: dict = strategy_config or api.get_tentacle_config(cls)
        if constants.STRATEGIES_REQUIRED_EVALUATORS in strategy_config:
            return strategy_config[constants.STRATEGIES_REQUIRED_EVALUATORS]
        else:
            raise Exception(f"'{constants.STRATEGIES_REQUIRED_EVALUATORS}' is missing in configuration file")

    @classmethod
    def get_compatible_evaluators_types(cls, strategy_config: dict = None) -> list:
        """
        :param strategy_config: the strategy configuration dict
        :return: the list of compatible evaluator type, [CONFIG_WILDCARD] means any type
        """
        strategy_config: dict = strategy_config or api.get_tentacle_config(cls)
        if constants.STRATEGIES_COMPATIBLE_EVALUATOR_TYPES in strategy_config:
            return strategy_config[constants.STRATEGIES_COMPATIBLE_EVALUATOR_TYPES]
        return [common_constants.CONFIG_WILDCARD]

    @classmethod
    def get_default_evaluators(cls, strategy_config: dict = None):
        strategy_config: dict = strategy_config or api.get_tentacle_config(cls)
        if constants.TENTACLE_DEFAULT_CONFIG in strategy_config:
            return strategy_config[constants.TENTACLE_DEFAULT_CONFIG]
        else:
            required_evaluators = cls.get_required_evaluators(strategy_config)
            if required_evaluators == common_constants.CONFIG_WILDCARD:
                return []
            return required_evaluators



#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import octobot_commons.tentacles_management as tentacles_management
import octobot_commons.constants as common_constants
import octobot_commons.logging as logging

import octobot_evaluators.api as api
import octobot_evaluators.evaluators as evaluator
import octobot_evaluators.constants as constants

LOGGER_NAME = "EvaluatorsFactory"


async def create_evaluators(evaluator_parent_class,
                            tentacles_setup_config: object,
                            matrix_id: str,
                            exchange_name: str,
                            bot_id: str,
                            crypto_currency_name_by_crypto_currencies: dict,
                            symbols_by_crypto_currency_tickers: dict,
                            symbols: list = None,
                            time_frames: list = None,
                            real_time_time_frames: list = None,
                            relevant_evaluators=common_constants.CONFIG_WILDCARD, ) -> list:
    return [
        await create_evaluator(evaluator_class,
                               tentacles_setup_config,
                               matrix_id=matrix_id,
                               exchange_name=exchange_name,
                               bot_id=bot_id,
                               cryptocurrency=cryptocurrency,
                               cryptocurrency_name=_get_cryptocurrency_name(
                                   evaluator_class,
                                   crypto_currency_name_by_crypto_currencies,
                                   cryptocurrency),
                               symbol=symbol,
                               time_frame=time_frame,
                               relevant_evaluators=relevant_evaluators,
                               all_symbols_by_crypto_currencies=symbols_by_crypto_currency_tickers,
                               time_frames=time_frames,
                               real_time_time_frames=real_time_time_frames
                               )
        for evaluator_class in tentacles_management.get_all_classes_from_parent(evaluator_parent_class)
        for cryptocurrency in _get_cryptocurrencies_to_create(evaluator_class,
                                                              crypto_currency_name_by_crypto_currencies)
        for symbol in _get_symbols_to_create(evaluator_class,
                                             symbols_by_crypto_currency_tickers,
                                             cryptocurrency,
                                             symbols)
        for time_frame in _get_time_frames_to_create(evaluator_class, time_frames)
    ]


def _get_cryptocurrency_name(evaluator_class, crypto_currency_name_by_crypto_currencies, cryptocurrency):
    return crypto_currency_name_by_crypto_currencies[cryptocurrency] \
        if crypto_currency_name_by_crypto_currencies \
           and cryptocurrency is not None \
           and not evaluator_class.get_is_cryptocurrency_name_wildcard() \
        else None


def _get_cryptocurrencies_to_create(evaluator_class, crypto_currency_name_by_crypto_currencies):
    return list(crypto_currency_name_by_crypto_currencies) \
        if crypto_currency_name_by_crypto_currencies and \
           not evaluator_class.get_is_cryptocurrencies_wildcard() else [None]


def _get_symbols_to_create(evaluator_class, symbols_by_crypto_currencies, cryptocurrency, symbols):
    currency_symbols = symbols
    if cryptocurrency is not None:
        currency_symbols = symbols_by_crypto_currencies.get(cryptocurrency, [])
    return currency_symbols if currency_symbols and not evaluator_class.get_is_symbol_wildcard() else [None]


def _get_time_frames_to_create(evaluator_class, time_frames):
    return time_frames if time_frames and not evaluator_class.get_is_time_frame_wildcard() else [None]


async def create_evaluator(evaluator_class,
                           tentacles_setup_config: object,
                           matrix_id: str,
                           exchange_name: str,
                           bot_id: str,
                           cryptocurrency: str = None,
                           cryptocurrency_name: str = None,
                           symbol: str = None,
                           time_frame=None,
                           relevant_evaluators=common_constants.CONFIG_WILDCARD,
                           all_symbols_by_crypto_currencies=None,
                           time_frames=None,
                           real_time_time_frames=None):
    try:
        eval_class_instance = evaluator_class()
        eval_class_instance.set_tentacles_setup_config(tentacles_setup_config)
        if api.is_relevant_evaluator(eval_class_instance, relevant_evaluators):
            eval_class_instance.logger = logging.get_logger(evaluator_class.get_name())
            eval_class_instance.matrix_id = matrix_id
            eval_class_instance.exchange_name = exchange_name if exchange_name else None
            eval_class_instance.cryptocurrency = cryptocurrency
            eval_class_instance.cryptocurrency_name = cryptocurrency_name
            eval_class_instance.symbol = symbol if symbol else None
            eval_class_instance.time_frame = time_frame if time_frame else eval_class_instance.time_frame
            eval_class_instance.evaluator_type = evaluator.evaluator_class_str_to_matrix_type_dict[
                eval_class_instance.__class__.mro()[constants.EVALUATOR_CLASS_TYPE_MRO_INDEX].__name__]
            eval_class_instance.initialize(all_symbols_by_crypto_currencies, time_frames, real_time_time_frames)
            await eval_class_instance.prepare()
            # handle backtesting
            await eval_class_instance.start_evaluator(bot_id)
            return eval_class_instance
    except Exception as e:
        logging.get_logger(LOGGER_NAME).exception(e, True, f"Error when creating evaluator {evaluator_class}: {e}")
    return None


async def create_all_type_evaluators(tentacles_setup_config: object,
                                     matrix_id: str,
                                     exchange_name: str,
                                     bot_id: str,
                                     symbols_by_crypto_currencies: dict = None,
                                     symbols: list = None,
                                     time_frames: list = None,
                                     real_time_time_frames: list = None,
                                     relevant_evaluators=common_constants.CONFIG_WILDCARD,
                                     ) -> list:
    if not api.get_activated_strategies_classes(tentacles_setup_config):
        # If no strategy is activated, there is no evaluator to create (their evaluation would not be used)
        logging.get_logger(LOGGER_NAME).info(
            f"No evaluator to create for {exchange_name}: no activated evaluator strategy.")
        return []
    crypto_currency_name_by_crypto_currencies = {}
    symbols_by_crypto_currency_tickers = {}
    try:
        import octobot_trading.api as exchange_api
        exchange_id = exchange_api.get_exchange_id_from_matrix_id(exchange_name, matrix_id)
        exchange_manager = exchange_api.get_exchange_manager_from_exchange_name_and_id(exchange_name, exchange_id)
        for name, symbol_list in symbols_by_crypto_currencies.items():
            if symbol_list:
                base = exchange_api.get_base_currency(exchange_manager, symbol_list[0])
                crypto_currency_name_by_crypto_currencies[base] = \
                    crypto_currency_name_by_crypto_currencies.get(base, "") + name
                symbols_by_crypto_currency_tickers[base] = \
                    symbols_by_crypto_currency_tickers.get(base, []) + symbol_list
        return [
            await create_evaluators(
                evaluator_type, tentacles_setup_config,
                matrix_id=matrix_id, exchange_name=exchange_name,
                bot_id=bot_id,
                crypto_currency_name_by_crypto_currencies=crypto_currency_name_by_crypto_currencies,
                symbols_by_crypto_currency_tickers=symbols_by_crypto_currency_tickers,
                symbols=symbols, time_frames=time_frames,
                real_time_time_frames=real_time_time_frames,
                relevant_evaluators=relevant_evaluators)
            for evaluator_type in evaluator.EvaluatorClassTypes.values()]
    except ImportError:
        logging.get_logger(LOGGER_NAME).error("create_evaluators requires Octobot-Trading package installed")
    return []



#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_evaluators.api import inspection
from octobot_evaluators.api import evaluators
from octobot_evaluators.api import initialization
from octobot_evaluators.api import matrix

from octobot_evaluators.api.inspection import (
    is_relevant_evaluator,
    get_relevant_TAs_for_strategy,
)
from octobot_evaluators.api.evaluators import (
    get_evaluator_classes_from_type,
    get_evaluators_time_frames,
    create_matrix,
    create_evaluators,
    stop_evaluator,
    stop_evaluator_channel,
    stop_all_evaluator_channels,
    create_evaluator,
    initialize_evaluators,
    create_all_type_evaluators,
)
from octobot_evaluators.api.initialization import (
    init_time_frames_from_strategies,
    get_activated_strategies_classes,
    del_evaluator_channels,
    matrix_channel_exists,
    create_evaluator_channels,
)
from octobot_evaluators.api.matrix import (
    get_matrix,
    del_matrix,
    get_node_children_by_names,
    get_children_list,
    has_children,
    get_value,
    get_type,
    get_time,
)

__all__ = [
    "is_relevant_evaluator",
    "get_relevant_TAs_for_strategy",
    "get_evaluator_classes_from_type",
    "get_evaluators_time_frames",
    "create_matrix",
    "create_evaluators",
    "stop_evaluator",
    "stop_evaluator_channel",
    "stop_all_evaluator_channels",
    "create_evaluator",
    "initialize_evaluators",
    "create_all_type_evaluators",
    "init_time_frames_from_strategies",
    "get_activated_strategies_classes",
    "del_evaluator_channels",
    "matrix_channel_exists",
    "create_evaluator_channels",
    "get_matrix",
    "del_matrix",
    "get_node_children_by_names",
    "get_children_list",
    "has_children",
    "get_value",
    "get_type",
    "get_time",
]




#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import copy

import async_channel.channels as channel_instances

import octobot_commons.constants as common_constants
import octobot_commons.logging as logging
import octobot_commons.tentacles_management as tentacles_management
import octobot_commons.time_frame_manager as time_frame_manager

import octobot_evaluators.api as api
import octobot_evaluators.evaluators.channel as evaluator_channels
import octobot_evaluators.matrix as matrix
import octobot_evaluators.evaluators as evaluator

LOGGER_NAME = "EvaluatorsAPI"


async def create_all_type_evaluators(tentacles_setup_config: object,
                                     matrix_id: str,
                                     exchange_name: str,
                                     bot_id: str,
                                     symbols_by_crypto_currencies: dict = None,
                                     symbols: list = None,
                                     time_frames: list = None,
                                     real_time_time_frames: list = None,
                                     relevant_evaluators=common_constants.CONFIG_WILDCARD,
                                     ) -> list:
    return await evaluator.create_all_type_evaluators(tentacles_setup_config=tentacles_setup_config,
                                                      matrix_id=matrix_id,
                                                      exchange_name=exchange_name,
                                                      bot_id=bot_id,
                                                      symbols_by_crypto_currencies=symbols_by_crypto_currencies,
                                                      symbols=symbols,
                                                      time_frames=time_frames,
                                                      real_time_time_frames=real_time_time_frames,
                                                      relevant_evaluators=relevant_evaluators)


async def create_evaluators(evaluator_parent_class,
                            tentacles_setup_config: object,
                            matrix_id: str,
                            exchange_name: str,
                            bot_id: str,
                            crypto_currency_name_by_crypto_currencies: dict,
                            symbols_by_crypto_currency_tickers: dict,
                            symbols: list = None,
                            time_frames: list = None,
                            real_time_time_frames: list = None,
                            relevant_evaluators=common_constants.CONFIG_WILDCARD, ) -> list:
    return await evaluator.create_evaluators(
        evaluator_parent_class=evaluator_parent_class,
        tentacles_setup_config=tentacles_setup_config,
        matrix_id=matrix_id,
        exchange_name=exchange_name,
        bot_id=bot_id,
        crypto_currency_name_by_crypto_currencies=crypto_currency_name_by_crypto_currencies,
        symbols_by_crypto_currency_tickers=symbols_by_crypto_currency_tickers,
        symbols=symbols,
        time_frames=time_frames,
        real_time_time_frames=real_time_time_frames,
        relevant_evaluators=relevant_evaluators)


async def create_evaluator(evaluator_class,
                           tentacles_setup_config: object,
                           matrix_id: str,
                           exchange_name: str,
                           bot_id: str,
                           cryptocurrency: str = None,
                           cryptocurrency_name: str = None,
                           symbol: str = None,
                           time_frame=None,
                           relevant_evaluators=common_constants.CONFIG_WILDCARD,
                           all_symbols_by_crypto_currencies=None,
                           time_frames=None,
                           real_time_time_frames=None):
    return await evaluator.create_evaluator(evaluator_class=evaluator_class,
                                            tentacles_setup_config=tentacles_setup_config,
                                            matrix_id=matrix_id,
                                            exchange_name=exchange_name,
                                            bot_id=bot_id,
                                            cryptocurrency=cryptocurrency,
                                            cryptocurrency_name=cryptocurrency_name,
                                            symbol=symbol,
                                            time_frame=time_frame,
                                            relevant_evaluators=relevant_evaluators,
                                            all_symbols_by_crypto_currencies=all_symbols_by_crypto_currencies,
                                            time_frames=time_frames,
                                            real_time_time_frames=real_time_time_frames)


def get_evaluator_classes_from_type(evaluator_type, tentacles_setup_config, activated_only=True) -> list:
    if activated_only:
        return [cls for cls in tentacles_management.get_all_classes_from_parent(
            evaluator.EvaluatorClassTypes[evaluator_type]) if cls.is_enabled(tentacles_setup_config, False)]
    return tentacles_management.get_all_classes_from_parent(evaluator.EvaluatorClassTypes[evaluator_type])


async def initialize_evaluators(config, tentacles_setup_config) -> str:
    """
    :param config: bot config
    :param tentacles_setup_config: tentacles configuration
    :return: initialized matrix id
    """
    _init_time_frames(config, tentacles_setup_config)

    return create_matrix()


def get_evaluators_time_frames(config) -> list:
    return time_frame_manager.get_config_time_frame(config)


def _init_time_frames(config, tentacles_setup_config) -> list:
    # Init time frames using enabled strategies
    api.init_time_frames_from_strategies(config, tentacles_setup_config)
    time_frames = copy.copy(time_frame_manager.get_config_time_frame(config))
    return time_frames


def create_matrix() -> str:
    created_matrix: matrix.Matrix = matrix.Matrix()
    matrix.Matrices.instance().add_matrix(created_matrix)
    return created_matrix.matrix_id


async def stop_evaluator(evaluator) -> None:
    return await evaluator.stop()


async def stop_evaluator_channel(matrix_id, chan_name) -> None:
    try:
        await evaluator_channels.get_chan(chan_name, matrix_id).stop()
    except Exception as e:
        logging.get_logger(LOGGER_NAME).exception(e, True, f"Error when stopping evaluator channel {chan_name}: {e}")


async def stop_all_evaluator_channels(matrix_id) -> None:
    for channel in channel_instances.ChannelInstances.instance().channels[matrix_id]:
        await stop_evaluator_channel(matrix_id, channel)



# cython: language_level=3
#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import async_channel.constants as channel_constants

import octobot_commons.logging as logging

import octobot_evaluators.evaluators.channel as evaluator_channels


class EvaluatorsChannelConsumer(evaluator_channels.EvaluatorChannelConsumer):
    """
    EvaluatorChannelConsumer adapted for EvaluatorsChannel
    """


class EvaluatorsChannelProducer(evaluator_channels.EvaluatorChannelProducer):
    """
    EvaluatorChannelProducer adapted for EvaluatorsChannel
    """

    # noinspection PyMethodOverriding
    async def send(self,
                   matrix_id,
                   data,
                   evaluator_name=channel_constants.CHANNEL_WILDCARD,
                   evaluator_type=channel_constants.CHANNEL_WILDCARD,
                   exchange_name=channel_constants.CHANNEL_WILDCARD,
                   cryptocurrency=channel_constants.CHANNEL_WILDCARD,
                   symbol=channel_constants.CHANNEL_WILDCARD,
                   time_frame=channel_constants.CHANNEL_WILDCARD,
                   origin_consumer=None):
        for consumer in self.channel.get_filtered_consumers(matrix_id=matrix_id,
                                                            evaluator_name=evaluator_name,
                                                            evaluator_type=evaluator_type,
                                                            exchange_name=exchange_name,
                                                            cryptocurrency=cryptocurrency,
                                                            symbol=symbol,
                                                            time_frame=time_frame,
                                                            origin_consumer=origin_consumer):
            await consumer.queue.put({
                "matrix_id": matrix_id,
                "evaluator_name": evaluator_name,
                "evaluator_type": evaluator_type,
                "exchange_name": exchange_name,
                "cryptocurrency": cryptocurrency,
                "symbol": symbol,
                "time_frame": time_frame,
                "data": data
            })


class EvaluatorsChannel(evaluator_channels.EvaluatorChannel):
    PRODUCER_CLASS = EvaluatorsChannelProducer
    CONSUMER_CLASS = EvaluatorsChannelConsumer

    MATRIX_ID_KEY = "matrix_id"
    EVALUATOR_NAME_KEY = "evaluator_name"
    EVALUATOR_TYPE_KEY = "evaluator_type"
    EXCHANGE_NAME_KEY = "exchange_name"
    CRYPTOCURRENCY_KEY = "cryptocurrency"
    SYMBOL_KEY = "symbol"
    TIME_FRAME_KEY = "time_frame"

    def __init__(self, matrix_id):
        super().__init__(matrix_id)
        self.logger = logging.get_logger(f"{self.__class__.__name__}")

    # noinspection PyMethodOverriding
    async def new_consumer(self,
                           callback: object,
                           size: int = 0,
                           priority_level: int = evaluator_channels.EvaluatorChannel.DEFAULT_PRIORITY_LEVEL,
                           matrix_id: str = channel_constants.CHANNEL_WILDCARD,
                           evaluator_name: str = channel_constants.CHANNEL_WILDCARD,
                           evaluator_type: object = channel_constants.CHANNEL_WILDCARD,
                           exchange_name: str = channel_constants.CHANNEL_WILDCARD,
                           cryptocurrency: str = channel_constants.CHANNEL_WILDCARD,
                           symbol: str = channel_constants.CHANNEL_WILDCARD,
                           time_frame=channel_constants.CHANNEL_WILDCARD) -> EvaluatorsChannelConsumer:
        consumer = EvaluatorsChannelConsumer(callback, size=size, priority_level=priority_level)
        await self._add_new_consumer_and_run(consumer,
                                             matrix_id=matrix_id,
                                             evaluator_name=evaluator_name,
                                             evaluator_type=evaluator_type,
                                             exchange_name=exchange_name,
                                             cryptocurrency=cryptocurrency,
                                             symbol=symbol,
                                             time_frame=time_frame)
        return consumer

    def get_filtered_consumers(self,
                               matrix_id=channel_constants.CHANNEL_WILDCARD,
                               evaluator_name=channel_constants.CHANNEL_WILDCARD,
                               evaluator_type=channel_constants.CHANNEL_WILDCARD,
                               exchange_name=channel_constants.CHANNEL_WILDCARD,
                               cryptocurrency=channel_constants.CHANNEL_WILDCARD,
                               symbol=channel_constants.CHANNEL_WILDCARD,
                               time_frame=channel_constants.CHANNEL_WILDCARD,
                               origin_consumer=None):
        return self.get_consumer_from_filters({
            self.MATRIX_ID_KEY: matrix_id,
            self.EVALUATOR_NAME_KEY: evaluator_name,
            self.EVALUATOR_TYPE_KEY: evaluator_type,
            self.EXCHANGE_NAME_KEY: exchange_name,
            self.CRYPTOCURRENCY_KEY: cryptocurrency,
            self.SYMBOL_KEY: symbol,
            self.TIME_FRAME_KEY: time_frame,
        },
            origin_consumer=origin_consumer)

    async def _add_new_consumer_and_run(self, consumer,
                                        matrix_id=channel_constants.CHANNEL_WILDCARD,
                                        evaluator_name=channel_constants.CHANNEL_WILDCARD,
                                        evaluator_type=channel_constants.CHANNEL_WILDCARD,
                                        exchange_name=channel_constants.CHANNEL_WILDCARD,
                                        cryptocurrency=channel_constants.CHANNEL_WILDCARD,
                                        symbol=channel_constants.CHANNEL_WILDCARD,
                                        time_frame=channel_constants.CHANNEL_WILDCARD):
        consumer_filters: dict = {
            self.MATRIX_ID_KEY: matrix_id,
            self.EVALUATOR_NAME_KEY: evaluator_name,
            self.EVALUATOR_TYPE_KEY: evaluator_type,
            self.EXCHANGE_NAME_KEY: exchange_name,
            self.CRYPTOCURRENCY_KEY: cryptocurrency,
            self.SYMBOL_KEY: symbol,
            self.TIME_FRAME_KEY: time_frame
        }

        self.add_new_consumer(consumer, consumer_filters)
        await consumer.run()



#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import abc 

import async_channel.channels as channels

import octobot_tentacles_manager.api as api

import octobot_evaluators.evaluators as evaluator


class SocialEvaluator(evaluator.AbstractEvaluator):
    __metaclass__ = evaluator.AbstractEvaluator
    SERVICE_FEED_CLASS = None

    def __init__(self):
        super().__init__()
        self.load_config()
        self.exchange_id = None

    def load_config(self):
        # try with this class name
        self.specific_config = api.get_tentacle_config(self.__class__)
        if not self.specific_config:
            # if nothing in config, try with any super-class' config file
            for super_class in self.get_parent_evaluator_classes(SocialEvaluator):
                self.specific_config = api.get_tentacle_config(super_class)
                if self.specific_config:
                    return
        # set default config if nothing found
        if not self.specific_config:
            self.set_default_config()

    # Override if no service feed is required for a social evaluator
    async def start(self, bot_id: str) -> bool:
        """
        :return: success of the evaluator's start
        """
        if self.SERVICE_FEED_CLASS is None:
            self.logger.error("SERVICE_FEED_CLASS is required to use a service feed. Consumer can't start.")
        else:
            await super().start(bot_id)
            try:
                import octobot_services.api as service_api
                service_feed = service_api.get_service_feed(self.SERVICE_FEED_CLASS, bot_id)
                if service_feed is not None:
                    service_feed.update_feed_config(self.specific_config)
                    await channels.get_chan(service_feed.FEED_CHANNEL.get_name()).new_consumer(self._feed_callback)
                    # store exchange_id to use it later for evaluation timestamps
                    import octobot_trading.api as exchange_api
                    self.exchange_id = exchange_api.get_exchange_id_from_matrix_id(self.exchange_name, self.matrix_id)
                    return True
            except ImportError as e:
                self.logger.exception(e, True, "Can't start: requires OctoBot-Services and OctoBot-Trading "
                                               "package installed")
        return False

    def get_current_exchange_time(self):
        try:
            import octobot_trading.api as exchange_api
            if self.exchange_id is not None:
                return exchange_api.get_exchange_current_time(
                    exchange_api.get_exchange_manager_from_exchange_name_and_id(
                        self.exchange_name,
                        self.exchange_id
                    )
                )
        except ImportError:
            self.logger.error(f"Can't get current exchange time: requires OctoBot-Trading package installed")
        return None

    def _get_tentacle_registration_topic(self, all_symbols_by_crypto_currencies, time_frames, real_time_time_frames):
        currencies, _, _ = super()._get_tentacle_registration_topic(all_symbols_by_crypto_currencies,
                                                                    time_frames,
                                                                    real_time_time_frames)
        symbols = [self.symbol]
        to_handle_time_frames = [self.time_frame]
        # by default no symbol registration for social evaluators
        # by default no time frame re+gistration for social evaluators
        return currencies, symbols, to_handle_time_frames

    @abc.abstractmethod
    async def _feed_callback(self, *args):
        raise NotImplementedError("_feed_callback is not implemented")
